/*************************************************************
*
* Description:
*	AT91SAM7xxx - Example program using IAR, ICC, GCC or Keil
*
* Author[s]:
*	Frank Goetze
**************************************************************/

#include "../../inc/pC_OS.h"

#define REG8   (volatile U08 *)
#define REG16  (volatile U16 *)
#define REG32  (volatile U32 *)

typedef volatile U32 AT91_REG;              // all registers are correct 32bit aligned
typedef struct _AT91S_A32 {
    AT91_REG     s[32];                     // Source Mode Register & Source Vector Register
} *AT91S_A32;
typedef struct _AT91S_A04 {
    AT91_REG     s[4];                      // Endpoint Control and Status Register & Endpoint FIFO Data Register
} *AT91S_A04;

#define OSTimer_INT_PRIO      2                                       // INT prio of SYS-TC
#define AT91S_PIT_BASE        (0xFFFFFD30)
#define AT91S_PIT_PIMR        (*(REG32 (AT91S_PIT_BASE + 0x00)))      // Period Interval Mode Register
#define AT91S_PIT_PISR        (*(REG32 (AT91S_PIT_BASE + 0x04)))      // Period Interval Status Register
#define AT91S_PIT_PIVR        (*(REG32 (AT91S_PIT_BASE + 0x08)))      // Period Interval Value Register
#define AT91S_PIT_PIIR        (*(REG32 (AT91S_PIT_BASE + 0x0C)))      // Period Interval Image Register
#define AT91S_PIT_PITEN       ((U32) 0x1 << 24)                       // (PITC) Periodic Interval Timer Enabled
#define AT91S_PIT_PITIEN      ((U32) 0x1 << 25)                       // (PITC) Periodic Interval Timer Interrupt Enable
#define AT91S_AIC_BASE        (0xFFFFF000)
#define AT91S_AIC_SMR         ((AT91S_A32)(AT91S_AIC_BASE + 0x00))    // [32] Source Mode Register
#define AT91S_AIC_SVR         ((AT91S_A32)(AT91S_AIC_BASE + 0x80))    // [32] Source Vector Register
#define AT91S_AIC_IECR        (*(REG32 (AT91S_AIC_BASE + 0x120)))     // Interrupt Enable Command Register
#define AT91S_AIC_IDCR        (*(REG32 (AT91S_AIC_BASE + 0x124)))     // Interrupt Disable Command Register
#define AT91S_AIC_ICCR        (*(REG32 (AT91S_AIC_BASE + 0x128)))     // Interrupt Clear Command Register
#define AT91S_ID_SYS          1U                                      // System-Peripheral ID
#define AT91S_ID_PIO          2U                                      // Parallel IO Port ID
#define AT91S_AIC_SRCTYPE_INT_LEVEL_SENSITIVE      0x00000000U        // (AIC) Internal Sources Code Label Level Sensitive
#define AT91S_PIOA_BASE       (0xFFFFF400U)                           // only PIO-A in hardware
#define AT91S_PIO_PER         (*(REG32 (AT91S_PIOA_BASE + 0x00)))     // PIO Enable Register
#define AT91S_PIO_OER         (*(REG32 (AT91S_PIOA_BASE + 0x10)))     // Output Enable Register
#define AT91S_PIO_SODR        (*(REG32 (AT91S_PIOA_BASE + 0x30)))     // Set Output Data Register
#define AT91S_PIO_CODR        (*(REG32 (AT91S_PIOA_BASE + 0x34)))     // Clear Output Data Register
#define AT91S_PIO_OWER        (*(REG32 (AT91S_PIOA_BASE + 0xA0)))     // Output Write Enable Register
#define AT91S_PIO_PA00        (1U << 0)                               // Pin Controlled by pin PA00 (for AT91SAM7S-EX)
#define AT91S_PIO_PA01        (1U << 1)                               // Pin Controlled by pin PA01
#define AT91S_PIO_PA02        (1U << 2)                               // Pin Controlled by pin PA02
#define AT91S_PIO_PA11        (1U << 11)                              // Pin Controlled by pin PA11 (for AT91SAM7Sxx-DevMaster)
#define AT91S_PMC_BASE        (0xFFFFFC00U)
#define AT91S_PMC_PCER        (*(REG32 (AT91S_PMC_BASE + 0x10)))      // Peripheral Clock Enable Register

#define AT91S_PIO_LEDc        AT91S_PIO_PA11                          // Pin used for the control-LED (for AT91SAM7Sxx)
//#define AT91S_PIO_LEDc        AT91S_PIO_PA01                          // Pin used for the control-LED (for AT91SAM7S-EK / AT91SAM7SE-EK)

#define OS_SYSTEM_TICKS_DIV   (OS_TIMER_TICKS_PER_SEC / OS_SYSTEM_TICKS_PER_SEC)

#define	TASK_STK_SIZE         256
#define MEM_POOL_SIZE         1024

OS_STK_TYPE Stk1[TASK_STK_SIZE];
OS_STK_TYPE Stk2[TASK_STK_SIZE];
OS_STK_TYPE Stk3[TASK_STK_SIZE];
U08 OS_ALIGN_ATTR pool[MEM_POOL_SIZE];

U32 OS_SystemTicksCnt;

OS_MUX Mutex1;
OS_EVG Events1;
#if OS_TMR_EN
OS_SEM LedSem;
OS_TMR LedTimer;
OS_TMR ROnceTimer1;
OS_TMR ROnceTimer2;
#endif

void  Task1(void *Tdata);
void  Task2(void *Tdata);
void  Task3(void *Tdata);

int main(void)
{
    OS_Init();
    OS_MemoryInit((OS_MEM OS_HUGE *)(&pool[0]), MEM_POOL_SIZE);

    OS_TaskCreate(Task1, NULL, &Stk1, TASK_STK_SIZE, 11);
    OS_TaskCreate(Task2, NULL, &Stk2, TASK_STK_SIZE, 13);
    OS_TaskCreate(Task3, NULL, &Stk3, TASK_STK_SIZE, 12);

    OS_MutexCreate(&Mutex1, 10);
    OS_EvgInit(&Events1);
#if OS_TMR_EN
    OS_SemInit(&LedSem, 0);
#endif
    
    OS_Start();

    return(0);            // normally never used since "OS_Start()" do not return
}

void OSTickISR(void)
{
    volatile U32 status;

    status = AT91S_PIT_PIVR;                               // ack the interrupt
    AT91S_AIC_ICCR  = AT91S_ID_SYS;                        // clear interrupt
#if OS_TMR_EN
    OSTimerTick();                                         // call pC/OS-function (timers)
#endif
    if((++OS_SystemTicksCnt) == OS_SYSTEM_TICKS_DIV) {
        OS_SystemTicksCnt = 0;
        OSSystemTick();                                    // call pC/OS-function (delays / wait)
    }
}

void AppTickInit(void)
{
    U32  counts;

    OS_SystemTicksCnt = 0;

    AT91S_PMC_PCER = (1 << AT91S_ID_PIO);                  // enable clock on PIO-mode
    AT91S_PIO_PER  = AT91S_PIO_LEDc | AT91S_PIO_PA02;      // Set in PIO mode
    AT91S_PIO_OER  = AT91S_PIO_LEDc | AT91S_PIO_PA02;      // Configure in Output
    AT91S_PIO_OWER = AT91S_PIO_LEDc | AT91S_PIO_PA02;      // allow writing to this pins
    AT91S_PIO_CODR = AT91S_PIO_LEDc | AT91S_PIO_PA02;      // Clear pin to set the control-LED on (test)
    AT91S_PIO_SODR = AT91S_PIO_LEDc | AT91S_PIO_PA02;      // Set pin to clear the control-LED (test)

    AT91S_AIC_IDCR = (1 << AT91S_ID_SYS);                  // disable the interrupt
    AT91S_AIC_SVR->s[AT91S_ID_SYS] = (U32)(OSTickISR);     // set the ISR-function
    AT91S_AIC_SMR->s[AT91S_ID_SYS] = AT91S_AIC_SRCTYPE_INT_LEVEL_SENSITIVE | OSTimer_INT_PRIO  ;
    counts = (48000000 / 16 / OS_TIMER_TICKS_PER_SEC) - 1; // calculate counter for 48MHz using PLL
    AT91S_PIT_PIMR = AT91S_PIT_PITEN | AT91S_PIT_PITIEN | counts;
    AT91S_AIC_ICCR = (1 << AT91S_ID_SYS);                  // clear the interrupt
    AT91S_AIC_IECR = (1 << AT91S_ID_SYS);                  // enable the interrupt
}

void Task1(void *Tdata)
{
    U08 returnOk;
    U32 event, i;
    U08 OS_HUGE *ptr;

    if(Tdata != NULL) while(1);
    AppTickInit();
    ptr=NULL;
    returnOk=OS_MemAlloc((void OS_HUGE *)(&ptr), 30);      // memory-manager test
    returnOk=OS_MemFree((void **)&ptr);
    while(1) {
        i=0;
        while(i<2500) i++;                                 // no Task-switch to MUTEX
        returnOk=OS_MutexPend(&Mutex1, OS_SUSPEND);        // first ok. - secound suspending
        if(!returnOk) {
            OS_TimeDly(5);
            returnOk=OS_MutexPendAbbort(&Mutex1);
            OS_TimeDly(5);
            returnOk=OS_MutexPost(&Mutex1);
            OS_TimeDly(5);
            event = 0x10;
            returnOk=OS_EvgPend(&Events1, &event, OS_EVG_OR_C, OS_SUSPEND);
        }
    }
}


void Task2(void *Tdata)
{
    U08 returnOk;
    U32 i;

    Tdata = Tdata;
    while(1) {
        returnOk=OS_MutexPend(&Mutex1, OS_SUSPEND);        // suspending
        if(!returnOk) {
            i=0;
            while(i<2500) i++;                             // no Taskswitch to Task1 or Task3
            returnOk=OS_MutexPost(&Mutex1);
            OS_TimeDly(5);
            OS_EvgPost(&Events1, 0x10, OS_EVG_OR);
        }
    }
}

#if OS_TMR_EN
void ROnceTimer1CB(void *tdata)
{
    tdata = tdata;
    OS_TimerStop(&LedTimer);                  // stope LED-timer
}

void ROnceTimer2CB(void *tdata)
{
    tdata = tdata;
    OS_TimerStart(&LedTimer, 0);              // restart LED-timer with org time
    OS_TimerCreate(&ROnceTimer1, 500, ROnceTimer1CB, NULL, OS_TMR_ENABLE | OS_TMR_CLR);
    OS_TimerStart(&ROnceTimer2, 0);           // start both run-once (one auto-clear) timer again
}

void LedTimerCB(void *tdata)
{
    tdata = tdata;
    OS_SemPost(&LedSem);
}
#endif

void Task3(void *Tdata)
{
    U08  LEDs = 0xFE;                                      // first ON bit-0

    Tdata = Tdata;
#if OS_TMR_EN
    OS_TimerCreate(&ROnceTimer1, 500, ROnceTimer1CB, NULL, OS_TMR_ENABLE | OS_TMR_CLR);
    OS_TimerCreate(&ROnceTimer2, 1000, ROnceTimer2CB, NULL, OS_TMR_ENABLE | OS_TMR_RONCE);
    OS_TimerCreate(&LedTimer, 30, LedTimerCB, NULL, OS_TMR_ENABLE | OS_TMR_CYCL);
#endif
    while(1) {
#if OS_TMR_EN
        OS_SemPend(&LedSem, OS_SUSPEND);                   // wait on timmer-triggert signal
#else
        OS_TimeDly(3);                                     // test delay-wakeup
#endif
        LEDs ^= 0x01;                                      // Toggle LED
        if(LEDs &0x01) {
            AT91S_PIO_CODR = AT91S_PIO_LEDc;               // Clear pin to set the control-LED on
            AT91S_PIO_SODR = AT91S_PIO_PA02;
        } else {
            AT91S_PIO_SODR = AT91S_PIO_LEDc;               // Set pin to clear the control-LED
            AT91S_PIO_CODR = AT91S_PIO_PA02;
        }
    }
}

