/***********************************************************
*                       pC/OS 1.95c                        *
*                   a FREE - RTOS Kernel                   *
*                                                          *
*                 based on the C/OS 1.00                  *
*                   by  Jean J. Labrosse                   *
*                                                          *
*      changed by Frank Goetze  -  www.embedded-os.de      *
************************************************************
* Permission to use, copy, modify, and distribute this     *
* software in source and binary forms and its              *
* documentation for any purpose and without fee is hereby  *
* granted, provided that the above authors notice appear   *
* in all copies and that both that authors notice and this *
* permission notice appear in supporting documentation.    *
*                                                          *
* THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND   *
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT    *
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY    *
* AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.     *
* IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT,  *
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR             *
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,    *
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF     *
* USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER *
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN        *
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING           *
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE   *
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY *
* OF SUCH DAMAGE.                                          *
************************************************************
*                      pC_OS_evg.c                         *
*                   EVENTGROUP SERVICE                     *
***********************************************************/

#if OS_EVG_EN
/*
************************************************************
*                   INITIALIZE EVENTGROUP                  *
************************************************************
*/
U08 OS_EvgInit(OS_EVG *pevg) KC51_REENTRANT
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    U08 i;

    OS_ENTER_CRITICAL();
    if (pevg) {
        pevg->OSEvents      = 0;
        pevg->OSEvgSuspList = (OS_EVG_SUSP OS_HUGE *)NULL;
        pevg->OSEvgGrp      = 0x00;
        for (i = 0; i < OS_PRIO_TBL; i++)
            pevg->OSEvgTbl[i] = 0x00;
        OS_EXIT_CRITICAL();
        return (OS_NO_ERR);
    } else {
        OS_EXIT_CRITICAL();
        return (OS_EVG_ERR);
    }
}
/*
************************************************************
*         LINK SUSPENDED ON EVENTGROUP (internal)          *
************************************************************
*/
void OSEvgLink(OS_EVG *pevg, OS_EVG_SUSP OS_HUGE *sEntry) KC51_REENTRANT
{
    OS_EVG_SUSP OS_HUGE *suspEntry;
    OS_EVG_SUSP OS_HUGE *prevEntry;

    prevEntry = (OS_EVG_SUSP OS_HUGE *)NULL;
    if ((suspEntry = pevg->OSEvgSuspList) == (OS_EVG_SUSP OS_HUGE *)NULL) {
        pevg->OSEvgSuspList = sEntry;
        return;
    }
    while ((suspEntry->OSsuspPrio < sEntry->OSsuspPrio) && (suspEntry->OSsuspNext != (OS_EVG_SUSP OS_HUGE *)NULL)) {
        prevEntry = suspEntry;
        suspEntry = suspEntry->OSsuspNext;
    }
    if (suspEntry->OSsuspPrio < sEntry->OSsuspPrio) {
        suspEntry->OSsuspNext = sEntry;
        return;
    }
    if (prevEntry != (OS_EVG_SUSP OS_HUGE *)NULL) {
        prevEntry->OSsuspNext = sEntry;
    } else {
        pevg->OSEvgSuspList = sEntry;
    }
    sEntry->OSsuspNext = suspEntry;
}
/*
************************************************************
*        UNLINK SUSPENDED FROM EVENTGROUP (internal)       *
************************************************************
*/
void OSEvgUnLink(OS_EVG *pevg, U08 prio) KC51_REENTRANT
{
    OS_EVG_SUSP OS_HUGE *suspEntry;
    OS_EVG_SUSP OS_HUGE *prevEntry;

    prevEntry = (OS_EVG_SUSP OS_HUGE *)NULL;
    if ((suspEntry = pevg->OSEvgSuspList) == (OS_EVG_SUSP OS_HUGE *)NULL) 
        return;
    while((suspEntry->OSsuspPrio != prio) &&  (suspEntry->OSsuspNext != (OS_EVG_SUSP OS_HUGE *)NULL)) {
        prevEntry = suspEntry;
        suspEntry = suspEntry->OSsuspNext;
    }
    if (suspEntry->OSsuspPrio != prio)
        return;
    if (prevEntry == (OS_EVG_SUSP OS_HUGE *)NULL)
        pevg->OSEvgSuspList   = suspEntry->OSsuspNext;
    else
        prevEntry->OSsuspNext = suspEntry->OSsuspNext;
}
/*
************************************************************
*       CHECK REQUIRED FLAGS ON EVENTGROUP (internal)      *
************************************************************
*/
U32 OSEvgCompare(OS_EVG *pevg, U32 events, U08 mode) KC51_REENTRANT
{
    switch (mode & ~OS_EVG_CONSUME) {
        case OS_EVG_OR:
            return (pevg->OSEvents & events);
            // break;
        case OS_EVG_AND:
            if ((pevg->OSEvents & events) == events)
                return (events);
            break;
        default:
            break;
    }
    return (0);
}
/*
************************************************************
*                     PEND ON EVENTGROUP                   *
************************************************************
*/
U08 OS_EvgPend(OS_EVG *pevg, U32 *events, U08 mode, U16 timeout) KC51_REENTRANT
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    OS_EVG_SUSP  sEntry;
    U32          tmp;
    U08          x, y, bitx, bity;

    if (!(mode & (OS_EVG_OR | OS_EVG_AND)) || !(events))
        return (OS_EVG_ERR);
    OS_ENTER_CRITICAL();
    tmp = OSEvgCompare(pevg, *events, mode);
    if (tmp != 0) {
        if (mode & OS_EVG_CONSUME)
            pevg->OSEvents &= ~tmp;
        *events = tmp;
        OS_EXIT_CRITICAL();
        return (OS_NO_ERR);
    }
    if ((timeout == OS_NO_SUSP) || (OSIntNesting > 0)) {
        OS_EXIT_CRITICAL();
        return (OS_EVG_NOE);
    }
    sEntry.OSsuspNext  = (OS_EVG_SUSP OS_HUGE *)NULL;
    sEntry.OSsuspPrio  = OSTCBCur->OSTCBPrio;
    sEntry.OSreqEvents = events;
    sEntry.OSreqMode   = mode;
    OSLockNesting++;
    OS_EXIT_CRITICAL();
    OSEvgLink(pevg, (OS_EVG_SUSP OS_HUGE *)&sEntry);
    OS_ENTER_CRITICAL();
    if (OSLockNesting != 0)
        OSLockNesting--;
#if OS_TASK_DESTROY_EN
    OSTCBCur->OSTCBipc   = pevg;
#endif
    OSTCBCur->OSTCBStat |= OS_STAT_EVG;
    OSTCBCur->OSTCBDly   = timeout;
    y                    = OSTCBCur->OSTCBPrio >> 3;
    x                    = OSTCBCur->OSTCBPrio & 0x07;
    bity                 = OSMapTbl[y];
    bitx                 = OSMapTbl[x];
    if ((OSRdyTbl[y] &= ~(bitx)) == 0)
        OSRdyGrp &= ~(bity);
    pevg->OSEvgTbl[y] |= bitx;
    pevg->OSEvgGrp    |= bity;
    OS_EXIT_CRITICAL();
    OSSched();
    OS_ENTER_CRITICAL();
    if (OSTCBCur->OSTCBStat & OS_STAT_EVG) {
        OSLockNesting++;
        OS_EXIT_CRITICAL();
        OSEvgUnLink(pevg, OSTCBCur->OSTCBPrio);
        OS_ENTER_CRITICAL();
        if (OSLockNesting != 0)
            OSLockNesting--;
        if ((U08)(pevg->OSEvgTbl[y] &= ~(bitx)) == 0)
            pevg->OSEvgGrp &= ~(bity);
        OSTCBCur->OSTCBStat = OS_STAT_RDY;
        OS_EXIT_CRITICAL();
        return (OS_TIMEOUT);
    }
    OS_EXIT_CRITICAL();
    return (OS_NO_ERR);
}
/*
************************************************************
*                    POST TO A EVENTGROUP                  *
************************************************************
*/
U08 OS_EvgPost(OS_EVG *pevg, U32 events, U08 mode) KC51_REENTRANT
{
#if OS_CRITICAL_SR
    OS_STK_TYPE          OS_sr;
#endif
    OS_EVG_SUSP OS_HUGE *suspEntry;
    U32                  tmp;
    U08                  bitx, bity, p;

    OS_ENTER_CRITICAL();
    if (mode & OS_EVG_OR)
        pevg->OSEvents |= events;
    else {
        if (mode & OS_EVG_CLR)
            pevg->OSEvents &= ~(events);
        else {
            OS_EXIT_CRITICAL();
            return (OS_EVG_ERR);
        }
    }
    if ((suspEntry = pevg->OSEvgSuspList) == (OS_EVG_SUSP OS_HUGE *)NULL) {
        OS_EXIT_CRITICAL();
        return (OS_NO_ERR);
    }    
    if (OSIntNesting > 0) {
        OS_EXIT_CRITICAL();
        return (OS_EVG_ERR);
    }
    do {
        tmp = OSEvgCompare(pevg, *suspEntry->OSreqEvents, suspEntry->OSreqMode);
        if (tmp != 0) {
            if (suspEntry->OSreqMode & OS_EVG_CONSUME)
                pevg->OSEvents &= ~tmp;
            *suspEntry->OSreqEvents = tmp;
            p = suspEntry->OSsuspPrio;
            OSLockNesting++;
            OS_EXIT_CRITICAL();
            OSEvgUnLink(pevg, p);
            OS_ENTER_CRITICAL();
            if (OSLockNesting != 0)
                OSLockNesting--;
            bity = OSMapTbl[p >> 3];
            bitx = OSMapTbl[p & 0x07];
            if ((U08)(pevg->OSEvgTbl[p >> 3] &= ~(bitx)) == 0)
                pevg->OSEvgGrp &= ~(bity);
            OSTCBPrioTbl[p]->OSTCBDly   = 0;
            OSTCBPrioTbl[p]->OSTCBStat &= ~(OS_STAT_EVG);
            OSRdyGrp                   |= bity;
            OSRdyTbl[p >> 3]           |= bitx;
        }
        suspEntry = suspEntry->OSsuspNext;
    } while (suspEntry != (OS_EVG_SUSP OS_HUGE *)NULL);
    OS_EXIT_CRITICAL();
    OSSched();
    return (OS_NO_ERR);
}
/*
************************************************************
*           ABBORT HIGHEST PENDER ON A EVENTGROUP          *
************************************************************
*/
U08 OS_EvgPendAbbort(OS_EVG *pevg) KC51_REENTRANT
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    U08          x, y, bitx, bity, p;

    OS_ENTER_CRITICAL();
    if (pevg->OSEvgGrp != 0) {
        y    = OSUnMapTbl[pevg->OSEvgGrp];
        x    = OSUnMapTbl[pevg->OSEvgTbl[y]];
        bity = OSMapTbl[y];
        bitx = OSMapTbl[x];
        if ((U08)(pevg->OSEvgTbl[y] &= ~(bitx)) == 0)
            pevg->OSEvgGrp &= ~(bity);
        p    = (y << 3) + x;
        OSLockNesting++;
        OS_EXIT_CRITICAL();
        OSEvgUnLink(pevg, p);
        OS_ENTER_CRITICAL();
        if (OSLockNesting != 0)
            OSLockNesting--;
        OSTCBPrioTbl[p]->OSTCBDly   = 0;
        OSTCBPrioTbl[p]->OSTCBStat &= ~(OS_STAT_EVG);
        OSRdyGrp                   |= bity;
        OSRdyTbl[y]                |= bitx;
        OS_EXIT_CRITICAL();
        OSSched();
        return (OS_NO_ERR);
    }
    OS_EXIT_CRITICAL();
    return (OS_TIME_NOT_DLY);
}

#if OS_TASK_DESTROY_EN
/*
************************************************************
*              DELETE A TASK FROM A EVENTGROUP             *
************************************************************
*/
void OSEvgDestroyTask(OS_EVG *pevg, U08 prio)
{
    U08          x, y, bitx, bity;

    y      = prio >> 3;
    x      = prio & 0x07;
    bity   = OSMapTbl[y];
    bitx   = OSMapTbl[x];
    if ((U08)(pevg->OSEvgTbl[y] &= ~(bitx)) == 0)
        pevg->OSEvgGrp &= ~(bity);
    OSEvgUnLink(pevg, prio);
    OSTCBPrioTbl[prio]->OSTCBDly   =  0;
    OSTCBPrioTbl[prio]->OSTCBStat &= ~(OS_STAT_EVG);
    OSRdyGrp                      |=  bity;
    OSRdyTbl[y]                   |=  bitx;
}
#endif /* OS_TASK_DESTROY_EN */
#endif /* OS_EVG_EN */

/************************* END ****************************/


